@@ -24,10 +24,11 @@ import papyon
 import papyon.event
 
 from butterfly.util.decorator import async
-from butterfly.handle import ButterflyHandleFactory
+from butterfly.channel import ButterflyChannel
 
 __all__ = ['ButterflyContactListChannelFactory']
 
+logger = logging.getLogger('Butterfly.ContactListChannel')
 
 class HandleMutex(object):
     def __init__(self):
@@ -95,22 +96,26 @@ def ButterflyContactListChannelFactory(connection, manager, handle, props):
         props[telepathy.CHANNEL_INTERFACE + '.TargetHandleType'],
         props[telepathy.CHANNEL_INTERFACE + '.TargetHandle'])
 
-    if handle.get_name() == 'subscribe':
+    if handle.get_name() == 'stored':
+        raise telepathy.errors.NotImplemented
+    elif handle.get_name() == 'subscribe':
         channel_class = ButterflySubscribeListChannel
     elif handle.get_name() == 'publish':
         channel_class = ButterflyPublishListChannel
     elif handle.get_name() == 'hide':
-        channel_class = ButterflyHideListChannel
+        raise telepathy.errors.NotImplemented
     elif handle.get_name() == 'allow':
-        channel_class = ButterflyAllowListChannel
+        raise telepathy.errors.NotImplemented
     elif handle.get_name() == 'deny':
-        channel_class = ButterflyDenyListChannel
+        raise telepathy.errors.NotImplemented
     else:
-        raise TypeError("Unknown list type : " + handle.get_name())
+        logger.error("Unknown list type : " + handle.get_name())
+        raise telepathy.errors.InvalidHandle
     return channel_class(connection, manager, props)
 
 
 class ButterflyListChannel(
+        ButterflyChannel,
         telepathy.server.ChannelTypeContactList,
         telepathy.server.ChannelInterfaceGroup,
         papyon.event.AddressBookEventInterface):
@@ -120,6 +125,7 @@ class ButterflyListChannel(
         self._conn_ref = weakref.ref(connection)
         telepathy.server.ChannelTypeContactList.__init__(self, connection, manager, props,
             object_path=object_path)
+        ButterflyChannel.__init__(self, connection, props)
         telepathy.server.ChannelInterfaceGroup.__init__(self)
         papyon.event.AddressBookEventInterface.__init__(self, connection.msn_client)
         self._populate(connection)
@@ -135,8 +141,7 @@ class ButterflyListChannel(
 
         ad, lp, rp = self._filter_contact(contact)
         if ad or lp or rp:
-            handle = ButterflyHandleFactory(self._conn_ref(), 'contact',
-                contact.account, contact.network_id)
+            handle = self._conn.ensure_contact_handle(contact)
             if ad: added.add(handle)
             if lp: local_pending.add(handle)
             if rp: remote_pending.add(handle)
@@ -146,8 +151,7 @@ class ButterflyListChannel(
 
     # papyon.event.AddressBookEventInterface
     def on_addressbook_contact_deleted(self, contact):
-        handle = ButterflyHandleFactory(self._conn_ref(), 'contact',
-                contact.account, contact.network_id)
+        handle = self._conn.ensure_contact_handle(contact)
         ad, lp, rp = self._filter_contact(contact)
         if self._contains_handle(handle) and not ad:
             self.MembersChanged('', (), [handle], (), (), 0,
@@ -170,8 +174,7 @@ class ButterflyListChannel(
         for contact in connection.msn_client.address_book.contacts:
             ad, lp, rp = self._filter_contact(contact)
             if ad or lp or rp:
-                handle = ButterflyHandleFactory(self._conn_ref(), 'contact',
-                        contact.account, contact.network_id)
+                handle = self._conn.ensure_contact_handle(contact)
                 if ad: added.add(handle)
                 if lp: local_pending.add(handle)
                 if rp: remote_pending.add(handle)
@@ -217,37 +220,87 @@ class ButterflySubscribeListChannel(ButterflyListChannel,
     @Lockable(mutex, 'add_subscribe', 'finished_cb')
     def _add(self, handle_id, message, finished_cb):
         handle = self._conn.handle(telepathy.HANDLE_TYPE_CONTACT, handle_id)
+        if handle is self._conn.GetSelfHandle():
+            return True # don't add ourself
         if handle.contact is not None and \
            handle.contact.is_member(papyon.Membership.FORWARD):
-            return True
+            return True # contact already there
 
         account = handle.account
         network = handle.network
         groups = list(handle.pending_groups)
         handle.pending_groups = set()
         ab = self._conn.msn_client.address_book
+
+        # We redefine these two callbacks for two reasons:
+        #
+        #  1. For failed, we can give a nice warning.
+        #
+        #  2. For both callbacks, it is more than likely that calling
+        #     finished_cb() will unlock the mutex and the publish
+        #     channel underneath's _add function will be called. If we
+        #     keep handle in scope, it won't be disposed and the
+        #     publish channel will have the same handle. This fixes a
+        #     few bugs:
+        #
+        #      I. When the contact doesn't actually exist, calling
+        #         finished_cb after the handle gets disposed means it
+        #         will no longer be in the connection's handle
+        #         dictionary and the publish channel's _add() will
+        #         raise InvalidHandle (see fd.o#27553). This isn't
+        #         actually a problem as the function should return
+        #         anyway but it doesn't get a chance to unlock the
+        #         mutex and it throws an unhandled exception which
+        #         apport users think is crazy.
+        #
+        #     II. Similar to above, but when the contact does
+        #         actually exist then the publish list wants to act
+        #         appropriately. If the handle has already been
+        #         disposed then it'll raise the same exception and
+        #         won't call the appropriate AB method.
+        #
+        #    III. When a contact is actually added, instead of the
+        #         handle already having been disposed of, it stays
+        #         around for a bit so you don't add handle n and then
+        #         MembersChanged gets fired for handle n+1.
+        #
+        # If these cases aren't in fact in play when this is called,
+        # then not only am I surprised, but no damage is done.
+
+        def failed_cb(error_code, *cb_args):
+            logger.warning('Failed to add messenger contact; '
+                           'error code: %s' % error_code)
+            finished_cb()
+            handle
+
+        def done_cb(*cb_args):
+            finished_cb()
+            handle
+
         ab.add_messenger_contact(account,
                 network_id=network,
                 auto_allow=False,
                 invite_message=message.encode('utf-8'),
                 groups=groups,
-                done_cb=(finished_cb,),
-                failed_cb=(finished_cb,))
+                done_cb=(done_cb,),
+                failed_cb=(failed_cb,))
 
     @Lockable(mutex, 'rem_subscribe', 'finished_cb')
     def _remove(self, handle_id, finished_cb):
         handle = self._conn.handle(telepathy.HANDLE_TYPE_CONTACT, handle_id)
         contact = handle.contact
+        if handle is self._conn.GetSelfHandle():
+            return True # don't try to remove ourself
         if contact is None or not contact.is_member(papyon.Membership.FORWARD):
-            return True
+            return True # contact not in address book
+
         ab = self._conn.msn_client.address_book
         ab.delete_contact(contact, done_cb=(finished_cb,),
                 failed_cb=(finished_cb,))
 
     # papyon.event.ContactEventInterface
     def on_contact_memberships_changed(self, contact):
-        handle = ButterflyHandleFactory(self._conn_ref(), 'contact',
-                contact.account, contact.network_id)
+        handle = self._conn.ensure_contact_handle(contact)
         if contact.is_member(papyon.Membership.FORWARD):
             self.MembersChanged('', [handle], (), (), (), 0,
                     telepathy.CHANNEL_GROUP_CHANGE_REASON_INVITED)
@@ -265,7 +318,8 @@ class ButterflyPublishListChannel(ButterflyListChannel,
         ButterflyListChannel.__init__(self, connection, manager, props,
             object_path='RosterChannel/List/publish')
         papyon.event.ContactEventInterface.__init__(self, connection.msn_client)
-        self.GroupFlagsChanged(0, 0)
+        self.GroupFlagsChanged(telepathy.CHANNEL_GROUP_FLAG_CAN_ADD |
+                telepathy.CHANNEL_GROUP_FLAG_CAN_REMOVE, 0)
 
     def AddMembers(self, contacts, message):
         for handle_id in contacts:
@@ -280,8 +334,7 @@ class ButterflyPublishListChannel(ButterflyListChannel,
         for contact in self._conn.msn_client.address_book.contacts:
             if not contact.is_member(papyon.Membership.PENDING):
                 continue
-            handle = ButterflyHandleFactory(self._conn_ref(), 'contact',
-                    contact.account, contact.network_id)
+            handle = self._conn.ensure_contact_handle(contact)
             result.append((handle, handle,
                     telepathy.CHANNEL_GROUP_CHANGE_REASON_INVITED,
                     contact.attributes.get('invite_message', '')))
@@ -296,8 +349,13 @@ class ButterflyPublishListChannel(ButterflyListChannel,
     def _add(self, handle_id, message, finished_cb):
         handle = self._conn.handle(telepathy.HANDLE_TYPE_CONTACT, handle_id)
         contact = handle.contact
+        if handle is self._conn.GetSelfHandle():
+            return True # don't add ourself
         if contact is not None and contact.is_member(papyon.Membership.ALLOW):
-            return True
+            return True # contact is already allowed
+        if contact is None:
+            logger.debug('Cannot allow/accept None contact %s' % handle.get_name())
+            return True # contact doesn't actually exist
 
         account = handle.account
         network = handle.network
@@ -314,6 +372,8 @@ class ButterflyPublishListChannel(ButterflyListChannel,
         handle = self._conn.handle(telepathy.HANDLE_TYPE_CONTACT, handle_id)
         contact = handle.contact
         ab = self._conn.msn_client.address_book
+        if handle is self._conn.GetSelfHandle():
+            return True # don't try to remove ourself
         if contact.is_member(papyon.Membership.PENDING):
             ab.decline_contact_invitation(contact, False, done_cb=(finished_cb,),
                     failed_cb=(finished_cb,))
@@ -321,12 +381,11 @@ class ButterflyPublishListChannel(ButterflyListChannel,
             ab.disallow_contact(contact, done_cb=(finished_cb,),
                     failed_cb=(finished_cb,))
         else:
-            return True
+            return True # contact is neither pending or allowed
 
     # papyon.event.ContactEventInterface
     def on_contact_memberships_changed(self, contact):
-        handle = ButterflyHandleFactory(self._conn_ref(), 'contact',
-                contact.account, contact.network_id)
+        handle = self._conn.ensure_contact_handle(contact)
         if self._contains_handle(handle):
             if contact.is_member(papyon.Membership.PENDING):
                 # Nothing worth our attention
