@@ -66,9 +66,9 @@ class ButterflyContacts(
                 logger.debug("Ignoring unsupported interface %s" % interface)
 
         handle_type = telepathy.HANDLE_TYPE_CONTACT
-        ret = {}
+        ret = dbus.Dictionary(signature='ua{sv}')
         for handle in handles:
-            ret[handle] = {}
+            ret[handle] = dbus.Dictionary(signature='sv')
 
         functions = {
             telepathy.CONNECTION :
