@@ -51,58 +51,42 @@ class ButterflyConnection(telepathy.server.Connection,
         papyon.event.InviteEventInterface,
         papyon.event.OfflineMessagesEventInterface):
 
-    _secret_parameters = set([
-            'password',
-            'http-proxy-password',
-            'https-proxy-password'
-            ])
-    _mandatory_parameters = {
-            'account' : 's',
-            'password' : 's'
-            }
-    _optional_parameters = {
-            'server' : 's',
-            'port' : 'q',
-            'http-proxy-server' : 's',
-            'http-proxy-port' : 'q',
-            'http-proxy-username' : 's',
-            'http-proxy-password' : 's',
-            'https-proxy-server' : 's',
-            'https-proxy-port' : 'q',
-            'https-proxy-username' : 's',
-            'https-proxy-password' : 's',
-            'http-method' : 'b',
-            }
-    _parameter_defaults = {
-            'server' : 'messenger.hotmail.com',
-            'port' : 1863,
-            'http-method' : False
-            }
 
-    def __init__(self, manager, parameters):
-        self.check_parameters(parameters)
+    def __init__(self, protocol, manager, parameters):
+        protocol.check_parameters(parameters)
 
         try:
             account = unicode(parameters['account'])
             self._server = (parameters['server'].encode('utf-8'), parameters['port'])
 
-            # Build the proxies configurations
             self._proxies = {}
+            self._direct_proxies = [None]
+            self._http_proxies = [None]
+            self._default_http_proxy = None
+            self._default_https_proxy = None
+
+            # Build the proxies configurations
+            self._try_http = parameters['http-method']
             proxy = build_proxy_infos(parameters, 'http')
             if proxy is not None:
-                self._proxies['http'] = proxy
+                self._http_proxies = [proxy]
+                self._default_http_proxy = proxy
             proxy = build_proxy_infos(parameters, 'https')
             if proxy is not None:
-                self._proxies['https'] = proxy
+                self._default_https_proxy = proxy
+
+            self._fill_suggested_proxies()
+            self._use_next_proxy()
 
             self._manager = weakref.proxy(manager)
-            self._new_client(use_http=parameters['http-method'])
+            self._new_client(use_http=self._try_http)
             self._account = (parameters['account'].encode('utf-8'),
                     parameters['password'].encode('utf-8'))
-            self._channel_manager = ButterflyChannelManager(self)
+            self._channel_manager = ButterflyChannelManager(self, protocol)
 
             # Call parent initializers
-            telepathy.server.Connection.__init__(self, 'msn', account, 'butterfly')
+            telepathy.server.Connection.__init__(self, 'msn', account,
+                    'butterfly', protocol)
             telepathy.server.ConnectionInterfaceRequests.__init__(self)
             ButterflyPresence.__init__(self)
             ButterflyAliasing.__init__(self)
@@ -111,14 +95,14 @@ class ButterflyConnection(telepathy.server.Connection,
             ButterflyContacts.__init__(self)
             ButterflyMailNotification.__init__(self)
 
-            self.set_self_handle(ButterflyHandleFactory(self, 'self'))
+            self_handle = self.create_handle(telepathy.HANDLE_TYPE_CONTACT,
+                    self._account[0])
+            self.set_self_handle(self_handle)
 
             self.__disconnect_reason = telepathy.CONNECTION_STATUS_REASON_NONE_SPECIFIED
             self._initial_presence = papyon.Presence.INVISIBLE
             self._initial_personal_message = None
 
-            self._manager.connected(self)
-
             logger.info("Connection to the account %s created" % account)
         except Exception, e:
             import traceback
@@ -127,21 +111,109 @@ class ButterflyConnection(telepathy.server.Connection,
 
     def _new_client(self, use_http=False):
         if hasattr(self, '_msn_client') and self._msn_client:
-            self._msn_client.logout()
+            if self._msn_client.state != papyon.event.ClientState.CLOSED:
+                self._msn_client.logout()
             self._msn_client._events_handlers.remove(self)
 
         if use_http:
             self._tried_http = True
             self._msn_client = papyon.Client(self._server, self._proxies,
-                papyon.transport.HTTPPollConnection)
+                papyon.transport.HTTPPollConnection, 18)
         else:
             self._tried_http = False
-            self._msn_client = papyon.Client(self._server, self._proxies)
+            self._msn_client = papyon.Client(self._server, self._proxies,
+                version=18)
 
         papyon.event.ClientEventInterface.__init__(self, self._msn_client)
         papyon.event.InviteEventInterface.__init__(self, self._msn_client)
         papyon.event.OfflineMessagesEventInterface.__init__(self, self._msn_client)
 
+    def _fill_suggested_proxies(self):
+        try:
+            import libproxy
+        except ImportError:
+            logger.warning("Please install libproxy python bindings to enable proxy support.")
+            return
+
+        factory = libproxy.ProxyFactory()
+
+        # Get DirectConnection proxies
+        proxies = factory.getProxies('none://messenger.msn.com:1863')
+        proxies = [self._parse_proxy(p) for p in proxies]
+        self._direct_proxies = proxies
+
+        # Get HTTP proxies (if not already set by a parameter)
+        if not self._default_http_proxy:
+            proxies = factory.getProxies('http://gateway.messenger.msn.com/')
+            proxies = [self._parse_proxy(p) for p in proxies]
+            self._http_proxies = proxies
+            if len(proxies) > 0:
+                self._default_http_proxy = proxies[0]
+
+        # Get HTTPS proxies (if not already set by a parameter)
+        if not self._default_https_proxy:
+            proxies = factory.getProxies('https://rsi.hotmail.com/rsi/rsi.asmx')
+            proxies = [self._parse_proxy(p) for p in proxies]
+            if len(proxies) > 0:
+                self._default_https_proxy = proxies[0]
+
+    def _use_next_proxy(self):
+        if not self._try_http and self._direct_proxies:
+            direct_proxy = self._direct_proxies.pop(0)
+            http_proxy   = self._default_http_proxy
+            https_proxy  = self._default_https_proxy
+        elif self._http_proxies:
+            self._try_http = True
+            direct_proxy = None
+            http_proxy   = self._http_proxies.pop(0)
+            https_proxy  = self._default_https_proxy
+        else:
+            return False
+
+        self._proxies['direct'] = direct_proxy
+        self._proxies['http']   = http_proxy
+        self._proxies['https']  = https_proxy
+
+        # Clean proxies (remove None)
+        for conn_type in ('direct', 'http', 'https'):
+            if conn_type not in self._proxies:
+                pass
+            proxy = self._proxies[conn_type]
+            if proxy is None:
+                del self._proxies[conn_type]
+            logger.info('Using %s proxy: %s' % (conn_type, proxy))
+
+        return True
+
+    def _parse_proxy(self, proxy):
+        # libproxy documentation states:
+        #
+        #  * The format of the returned proxy strings are as follows:
+        #  *   - http://[username:password@]proxy:port
+        #  *   - socks://[username:password@]proxy:port
+        #  *   - direct://
+        #  etc.
+
+        if proxy is None or proxy == "direct://":
+            return None
+
+        index = proxy.find("://")
+        ptype = proxy[0:index]
+        proxy = proxy[index + 3:]
+
+        # Get username and password out.
+        if '@' in proxy:
+            auth, proxy = proxy.split('@')
+            user, password = auth.split(':')
+        else:
+            user = password = None
+
+        if ':' in proxy:
+            server, port = proxy.split(':')
+
+        return papyon.ProxyInfos(host=server, port=int(port), type=ptype,
+                user=user, password=password)
+
     @property
     def manager(self):
         return self._manager
@@ -154,6 +226,44 @@ class ButterflyConnection(telepathy.server.Connection,
         self.check_handle(handle_type, handle_id)
         return self._handles[handle_type, handle_id]
 
+    def create_handle(self, handle_type, handle_name, **kwargs):
+        """Create new handle with given type and name."""
+        handle_id = self.get_handle_id()
+        handle = ButterflyHandleFactory(self, handle_type, handle_id,
+                handle_name, **kwargs)
+        if handle is None:
+            raise telepathy.NotAvailable('Handle type unsupported %d' % handle_type)
+        logger.info("New Handle %s" % unicode(handle))
+        self._handles[handle_type, handle_id] = handle
+        return handle
+
+    def is_valid_handle_name(self, handle_type, handle_name):
+        """Make sure the name is valid for this type of handle."""
+        if handle_type == telepathy.HANDLE_TYPE_CONTACT:
+            if '@' not in handle_name:
+                return False
+            if '.' not in handle_name.split("@", 1)[1]:
+                return False
+        return True
+
+    def normalize_handle_name(self, handle_type, handle_name):
+        """Normalize handle name so the name is consistent everywhere."""
+        if not self.is_valid_handle_name(handle_type, handle_name):
+            raise telepathy.InvalidHandle('TargetID %s not valid for type %d' %
+                (name, handle_type))
+        if handle_type == telepathy.HANDLE_TYPE_CONTACT:
+            return handle_name.lower().strip()
+        return handle_name
+
+    def ensure_contact_handle(self, contact):
+        """Build handle name for contact and ensure handle."""
+        if contact is None:
+            return telepathy.NoneHandler()
+        handle_type = telepathy.HANDLE_TYPE_CONTACT
+        extension = network_to_extension.get(contact.network_id, "")
+        handle_name = contact.account.lower() + extension
+        return self.ensure_handle(handle_type, handle_name, contact=contact)
+
     def Connect(self):
         if self._status == telepathy.CONNECTION_STATUS_DISCONNECTED:
             logger.info("Connecting")
@@ -163,7 +273,17 @@ class ButterflyConnection(telepathy.server.Connection,
     def Disconnect(self):
         logger.info("Disconnecting")
         self.__disconnect_reason = telepathy.CONNECTION_STATUS_REASON_REQUESTED
-        self._msn_client.logout()
+        if self._msn_client.state != papyon.event.ClientState.CLOSED:
+            self._msn_client.logout()
+        else:
+            self._disconnected()
+
+    def _disconnected(self):
+        logger.info("Disconnected")
+        self.StatusChanged(telepathy.CONNECTION_STATUS_DISCONNECTED,
+                self.__disconnect_reason)
+        self._channel_manager.close()
+        self._manager.disconnected(self)
 
     def GetInterfaces(self):
         # The self._interfaces set is only ever touched in ButterflyConnection.__init__,
@@ -179,34 +299,6 @@ class ButterflyConnection(telepathy.server.Connection,
 
         return self._interfaces
 
-    def RequestHandles(self, handle_type, names, sender):
-        self.check_connected()
-        self.check_handle_type(handle_type)
-
-        handles = []
-        for name in names:
-            if handle_type == telepathy.HANDLE_TYPE_CONTACT:
-                network_id = papyon.NetworkID.MSN
-                for network, extension in network_to_extension.items():
-                    if name.endswith(extension):
-                        name = name[0:-len(extension)]
-                        network_id = network
-                        break
-                contacts = self.msn_client.address_book.contacts.\
-                        search_by_account(name).\
-                        search_by_network_id(network_id)
-                handle = ButterflyHandleFactory(self, 'contact',
-                        name, network_id)
-            elif handle_type == telepathy.HANDLE_TYPE_LIST:
-                handle = ButterflyHandleFactory(self, 'list', name)
-            elif handle_type == telepathy.HANDLE_TYPE_GROUP:
-                handle = ButterflyHandleFactory(self, 'group', name)
-            else:
-                raise telepathy.NotAvailable('Handle type unsupported %d' % handle_type)
-            handles.append(handle.id)
-            self.add_client_handle(handle, sender)
-        return handles
-
     def _generate_props(self, channel_type, handle, suppress_handler, initiator_handle=None):
         props = {
             telepathy.CHANNEL_INTERFACE + '.ChannelType': channel_type,
@@ -248,33 +340,33 @@ class ButterflyConnection(telepathy.server.Connection,
             self.StatusChanged(telepathy.CONNECTION_STATUS_CONNECTING,
                     telepathy.CONNECTION_STATUS_REASON_REQUESTED)
         elif state == papyon.event.ClientState.SYNCHRONIZED:
-            handle = ButterflyHandleFactory(self, 'list', 'subscribe')
+            handle = self.ensure_handle(telepathy.HANDLE_TYPE_LIST, 'subscribe')
             props = self._generate_props(telepathy.CHANNEL_TYPE_CONTACT_LIST,
                 handle, False)
             self._channel_manager.channel_for_props(props, signal=True)
 
-            handle = ButterflyHandleFactory(self, 'list', 'publish')
+            handle = self.ensure_handle(telepathy.HANDLE_TYPE_LIST, 'publish')
             props = self._generate_props(telepathy.CHANNEL_TYPE_CONTACT_LIST,
                 handle, False)
             self._channel_manager.channel_for_props(props, signal=True)
 
-            #handle = ButterflyHandleFactory(self, 'list', 'hide')
+            #handle = self.ensure_handle(telepathy.HANDLE_TYPE_LIST, 'hide')
             #props = self._generate_props(telepathy.CHANNEL_TYPE_CONTACT_LIST,
             #    handle, False)
             #self._channel_manager.channel_for_props(props, signal=True)
 
-            #handle = ButterflyHandleFactory(self, 'list', 'allow')
+            #handle = self.ensure_handle(telepathy.HANDLE_TYPE_LIST, 'allow')
             #props = self._generate_propstelepathy.CHANNEL_TYPE_CONTACT_LIST,
             #    handle, False)
             #self._channel_manager.channel_for_props(props, signal=True)
 
-            #handle = ButterflyHandleFactory(self, 'list', 'deny')
+            #handle = self.ensure_handle(telepathy.HANDLE_TYPE_LIST, 'deny')
             #props = self._generate_props(telepathy.CHANNEL_TYPE_CONTACT_LIST,
             #    handle, False)
             #self._channel_manager.channel_for_props(props, signal=True)
 
             for group in self.msn_client.address_book.groups:
-                handle = ButterflyHandleFactory(self, 'group',
+                handle = self.ensure_handle(telepathy.HANDLE_TYPE_GROUP,
                         group.name.decode("utf-8"))
                 props = self._generate_props(
                     telepathy.CHANNEL_TYPE_CONTACT_LIST, handle, False)
@@ -294,22 +386,24 @@ class ButterflyConnection(telepathy.server.Connection,
             self._client.profile.end_point_name = "PAPYON"
 
             if (presence is not None) or (message is not None):
-                self._presence_changed(ButterflyHandleFactory(self, 'self'),
+                self._presence_changed(self._self_handle,
                         self._client.profile.presence,
                         self._client.profile.personal_message)
         elif state == papyon.event.ClientState.CLOSED:
-            self.StatusChanged(telepathy.CONNECTION_STATUS_DISCONNECTED,
-                    self.__disconnect_reason)
-            #FIXME
-            self._channel_manager.close()
-            self._advertise_disconnected()
+            self._disconnected()
 
     # papyon.event.ClientEventInterface
     def on_client_error(self, type, error):
         if type == papyon.event.ClientErrorType.NETWORK:
-            if self._tried_http is False:
-                logger.info("Failed to connect directly, trying HTTP")
-                self._new_client(use_http=True)
+            # Only move onto the next proxy if we've not already tried
+            # HTTP and we're in the connecting state. We don't want to
+            # connect to HTTP if we're already connected and we lose
+            # connectivity (see fd.o#26147).
+            if self._status == telepathy.CONNECTION_STATUS_CONNECTING and \
+                    self._use_next_proxy():
+                logger.info("Failed to connect, trying HTTP "
+                            "(possibly again with another proxy)")
+                self._new_client(use_http=self._try_http)
                 self._msn_client.login(*self._account)
             else:
                 self.__disconnect_reason = telepathy.CONNECTION_STATUS_REASON_NETWORK_ERROR
@@ -327,8 +421,7 @@ class ButterflyConnection(telepathy.server.Connection,
 
         if len(conversation.participants) == 1:
             p = list(conversation.participants)[0]
-            handle = ButterflyHandleFactory(self, 'contact',
-                    p.account, p.network_id)
+            handle = self.ensure_contact_handle(p)
         else:
             handle = telepathy.server.handle.NoneHandle()
 
@@ -344,29 +437,24 @@ class ButterflyConnection(telepathy.server.Connection,
 
     # papyon.event.InviteEventInterface
     def on_invite_conference(self, call):
-        logger.debug("Call invite, ignoring it")
-        #logger.debug("Call invite")
-        #handle = ButterflyHandleFactory(self, 'contact', call.peer.account,
-                #call.peer.network_id)
+        logger.debug("Call invite")
 
-        #props = self._generate_props(telepathy.CHANNEL_TYPE_STREAMED_MEDIA,
-                #handle, False, initiator_handle=handle)
-
-        #channel = self._channel_manager.channel_for_props(props,
-                #signal=True, call=call)
+        handle = self.ensure_contact_handle(call.peer)
+        props = self._generate_props(telepathy.CHANNEL_TYPE_STREAMED_MEDIA,
+                handle, False, initiator_handle=handle)
+        channel = self._channel_manager.channel_for_props(props,
+                signal=True, call=call)
 
     # papyon.event.InviteEventInterface
     def on_invite_webcam(self, session, producer):
         direction = (producer and "send") or "receive"
-        logger.debug("Invitation to %s webcam, ignoring it" % direction)
-        #logger.debug("Invitation to %s webcam" % direction)
+        logger.debug("Invitation to %s webcam" % direction)
 
-        #handle = ButterflyHandleFactory(self, 'contact', session.peer.account,
-                #session.peer.network_id)
-        #props = self._generate_props(telepathy.CHANNEL_TYPE_STREAMED_MEDIA,
-                #handle, False, initiator_handle=handle)
-        #channel = self._channel_manager.channel_for_props(props, signal=True,
-                #call=session)
+        handle = self.ensure_contact_handle(session.peer)
+        props = self._generate_props(telepathy.CHANNEL_TYPE_STREAMED_MEDIA,
+                handle, False, initiator_handle=handle)
+        channel = self._channel_manager.channel_for_props(props, signal=True,
+                call=session)
 
     # papyon.event.OfflineMessagesEventInterface
     def on_oim_messages_received(self, messages):
@@ -377,11 +465,13 @@ class ButterflyConnection(telepathy.server.Connection,
     # papyon.event.OfflineMessagesEventInterface
     def on_oim_messages_fetched(self, messages):
         for message in messages:
+            if message.sender is None:
+                continue
             # Request butterfly text channel (creation, what happen when it exist)
             sender = message.sender
             logger.info('received offline message from %s : %s' % (sender.account, message.text))
-            handle = ButterflyHandleFactory(self, 'contact',
-                    sender.account, sender.network_id)
+
+            handle = self.ensure_contact_handle(sender)
             props = self._generate_props(telepathy.CHANNEL_TYPE_TEXT,
                 handle, False)
             channel = self._channel_manager.channel_for_props(props,
@@ -389,11 +479,18 @@ class ButterflyConnection(telepathy.server.Connection,
             # Notify it of the message
             channel.offline_message_received(message)
 
-    def _advertise_disconnected(self):
-        self._manager.disconnected(self)
+    # papyon.event.InviteEventInterface
+    def on_invite_file_transfer(self, session):
+        logger.debug("File transfer invite")
+
+        handle = self.ensure_contact_handle(session.peer)
+        props = self._generate_props(telepathy.CHANNEL_TYPE_FILE_TRANSFER,
+                handle, False)
+        channel = self._channel_manager.create_channel_for_props(props,
+                signal=True, session=session)
 
 
-def build_proxy_infos(self, parameters, proxy_type='http'):
+def build_proxy_infos(parameters, proxy_type='http'):
     server_key = proxy_type + '-proxy-server'
     port_key = proxy_type + '-proxy-port'
     username_key = proxy_type + '-proxy-username'
